import numpy as np
import matplotlib.pyplot as plt

from typing import Iterable

import parse_results


def make_bars_h(df, h_vals: Iterable[int], metric: str, normalized_util: float, period_size: str, m: int, theta: float, q:float=None, ylim:float=None) -> None:
    '''
    Generates bar graph comparing pi-blocking between SMLP and OMLP for different values of \mathcal{H}.
    
    Args:
        df (pandas.DataFrame): A pandas DataFrame object containing all results, typically loaded using pandas.read_csv.
        h_vals (Iterable[int]): A list of values of \mathcal{H} to be selected and entered into the graph.
        metric (str): The metric to be graphed. Must be one of 'max', 'mean', or 'pct'.
        normalized_util (float): A value from [0.0, 1.0] corresponding to a normalized system utilization value.
        period_size (str): 'small', 'medium', or 'large'.
        m (int): Value of \mathcal{M}, the number of cores.
        theta (float): The value of \Theta.
        q (float): Required if metric is set to 'pct'. Set to get the qth quantile largest worst-case pi-blocking bound.
        ylim (float): Optional. Sets the y-axis limit displayed on the graph.
    '''
    protocols = ('SMLP', 'OMLP')
    colors = ('#8c82cc', '#ff8b91')
    b_dict = {}
    if not h_vals:
        raise ValueError('H value iterable cannot be empty.')
    for h in h_vals:
        df_smlp, df_omlp = parse_results.filter_results(df, normalized_util, period_size, m, h, theta)
        match metric:
            case 'max':
                smlp_b, omlp_b = parse_results.get_max_block(df_smlp, df_omlp)
                metric_pretty = 'Maximum'
                metric_short = 'max'
            case 'mean':
                smlp_b, omlp_b = parse_results.get_mean_block(df_smlp, df_omlp)
                metric_pretty = 'Mean'
                metric_short = 'mean'
            case 'pct':
                if q is None:
                    raise ValueError('When metric is set to \'pct\', a value for q must be provided.')
                elif q < 0 or q > 1:
                    raise ValueError('q must be a value in [0.0, 1.0]')
                else:
                    smlp_b, omlp_b = parse_results.get_pct_block(df_smlp, df_omlp, q)
                    metric_pretty = f'{round(q * 100, 3)}th percentile'
                    metric_short = ''
            case _:
                raise ValueError('metric arg must be one of \'max\', \'mean\', or \'pct\'.')
        b_dict[h] = (smlp_b, omlp_b)
    x = np.arange(len(h_vals))

    plt.style.use('default')
    fig, ax = plt.subplots(layout='constrained')
    fig.set_figwidth(5.25)
    width = 0.25
    multiplier = 0
    for i in range(len(protocols)):
        offset = width * multiplier

        rects = ax.bar(x + offset, [round(b_dict[h][i], 3) for h in h_vals], width, label=protocols[i], color=colors[i])
        ax.bar_label(rects, padding=3, size=8)
        multiplier += 1

    ax.tick_params(axis='both', which='major', labelsize=12)
    ax.set_xticks(x + width/2, (f'$\\mathcal{{H}} = {h}$' for h in h_vals), fontsize=14)
    ax.set_ylim(0, ylim)  # Set this appropriately plz
    ax.set_ylabel(f'{metric_short} pi-blocking duration (ms)', fontsize=14)
    ax.legend(loc='upper right', fontsize=14)
    ax.set_axisbelow(True)
    ax.grid(axis='y')

    plt.title(f'{metric_pretty} observed worst-case pi-blocking by values of $\\mathcal{{H}}$\n', fontsize=16)

    plt.savefig('pdf/exp_H_bars.pdf', bbox_inches='tight')

    return


def make_util_plot(df, metric: str, period_size: str, m: int, h: int, theta: float, q:float=None, ylim:float=None) -> None:
    '''
    Plots pi-blocking durations for SMLP and OMLP across different system utilization values.

    Args:
        df (pandas.DataFrame): A pandas DataFrame object containing all results, typically loaded using pandas.read_csv.
        metric (str): The metric to be graphed. Must be one of 'max', 'mean', or 'pct'.
        period_size (str): 'small', 'medium', or 'large'.
        m (int): Value of \mathcal{M}, the number of cores.
        h (int): Value of \mathcal{H}, the size of the unit by which SMs are allocated.
        theta (float): The value of \Theta.
        q (float): Required if metric is set to 'pct'. Set to get the qth quantile largest worst-case pi-blocking bound.
        ylim (float): Optional. Sets the y-axis limit displayed on the graph.
    '''
    utils = [round(0.1 * i, 1) for i in range(2, 10)]

    protocols = ('SMLP', 'OMLP')
    colors = ('#8c82cc', '#ff8b91')
    markers = ('s', 'o')

    plt.style.use('default')

    b_dict = {}

    match metric:
        case 'max':
            getter = parse_results.get_max_block
            metric_pretty = 'Maximum'
        case 'mean':
            getter = parse_results.get_mean_block
            metric_pretty = 'Mean'
        case 'pct':
            if q is None:
                raise ValueError('When metric is set to \'pct\', a value for q must be provided.')
            elif q < 0 or q > 1:
                raise ValueError('q must be a value in [0.0, 1.0]')
            else:
                getter = lambda s, o: parse_results.get_pct_block(s, o, q)
                metric_pretty = f'{round(q * 100, 3)}th percentile'
        case _:
            raise ValueError('metric arg must be one of \'max\', \'mean\', or \'pct\'.')
    
    for u in utils:
        df_smlp, df_omlp = parse_results.filter_results(df, u, period_size, m, h, theta)
        b_dict[u] = getter(df_smlp, df_omlp)
    
    fig, ax = plt.subplots(layout='constrained')

    for i in range(len(protocols)):
        y = [b_dict[u][i] for u in utils]
        ax.plot(utils, y, '-' + markers[i], label=protocols[i], color=colors[i], markerfacecolor=colors[i])
    
    ax.tick_params(axis='both', which='major', labelsize=12)
    ax.set_xticks(utils, labels=[f'{u:.1f}' for u in utils])
    ax.set_xlabel(f'Normalized system utilization, $\\mathcal{{M}} = {m}$', fontsize=14)
    ax.set_ylabel(f'{metric} pi-blocking duration (ms)', fontsize=14)
    ax.grid(axis='both')
    ax.legend(fontsize=14)

    plt.title(f'{metric_pretty} observed worst-case pi-blocking by system utilization\n', fontsize=16)

    plt.savefig('pdf/exp_util_plot.pdf', bbox_inches='tight')

    return
