import pandas as pd
import numpy as np

from typing import Tuple, List, Iterable


def filter_results(df: pd.DataFrame, normalized_util: float, period_size: str, m: int, h: int, theta: float) -> Tuple[pd.DataFrame, pd.DataFrame]:
    '''
    Returns two DataFrames, one each for SMLP and OMLP results in that order, matching the specified params.
    '''
    match period_size:
        case 'small':
            tmin, tmax = 3, 33
        case 'medium':
            tmin, tmax = 10, 100
        case 'large':
            tmin, tmax = 50, 200
        case _:
            raise ValueError('period_size arg must be one of \'small\', \'medium\', or \'large\'.')
        
    df_filtered = df.loc[(df['Tmin'] == tmin) & (df['u'] == normalized_util) & (df['M'] == m) & (df['H'] == h) & (df['Theta'] == theta)]
    df_smlp = df_filtered.loc[df_filtered['SMLP'] == 1]
    df_omlp = df_filtered.loc[df_filtered['SMLP'] == 0]

    return df_smlp, df_omlp


def get_max_block(df_smlp: pd.DataFrame, df_omlp: pd.DataFrame) -> Tuple[float, float]:
    '''
    Returns the maximum observed pi-blocking durations for the SMLP and OMLP protocols, in that order.
    '''
    smlp_b = df_smlp.loc[:,'wpiblock'].max()
    omlp_b = df_omlp.loc[:,'wpiblock'].max()

    return smlp_b, omlp_b


def get_mean_block(df_smlp: pd.DataFrame, df_omlp: pd.DataFrame) -> Tuple[float, float]:
    '''
    Returns the average pi-blocking durations for the SMLP and OMLP protocols, in that order.
    '''
    smlp_mean = df_smlp.loc[:,'wpiblock'].mean()
    omlp_mean = df_omlp.loc[:,'wpiblock'].mean()

    return smlp_mean, omlp_mean


def get_pct_block(df_smlp: pd.DataFrame, df_omlp: pd.DataFrame, q: float) -> Tuple[float, float]:
    '''
    Returns pi-blocking durations at quantile q for the SMLP and OMLP protocols, in that order.
    '''
    smlp_b = df_smlp.loc[:,'wpiblock'].quantile(q=q, interpolation='lower')
    omlp_b = df_omlp.loc[:,'wpiblock'].quantile(q=q, interpolation='lower')

    return smlp_b, omlp_b
